# Generated by Django 3.2.5 on 2023-04-04 11:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('bank_sync', '0016_ipn_statement_url'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ipn',
            name='statement_url',
            field=models.URLField(),
        ),
    ]
