# Generated by Django 3.2.5 on 2022-07-14 12:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('bank_sync', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='callbacks',
            name='reference',
            field=models.CharField(max_length=100, unique=True),
        ),
    ]
