# Generated by Django 3.2.5 on 2022-11-09 07:52

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('bank_sync', '0005_alter_requestlogs_error'),
    ]

    operations = [
        migrations.AddField(
            model_name='requestlogs',
            name='log_time',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]
